/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "server.h"
#include "util/memory.h"
#include "assert/assert.h"
#include "console/console.h"

server_t SERVER;

void serverInit() {
  memoryZero(&SERVER, sizeof(server_t));
  SERVER.state = SERVER_STATE_STOPPED;
}

errorret_t serverStart(const serverstart_t start) {
  assertIsMainThread("Server start must be on main thread");

  // Do not start a running server.
  if(SERVER.state != SERVER_STATE_STOPPED) {
    return error("Server is already running");
  }

  SERVER.type = start.type;

  // Hand off to relevant server type to start.
  switch(start.type) {
    case SERVER_TYPE_NETWORKED:
      return networkedServerStart(&SERVER, start);
      break;

    default:
      assertUnreachable("Invalid server type");
  }
}

uint8_t serverGetClientCount() {
  uint8_t i = 0, clientCount = 0;
  do {
    if(SERVER.clients[i].state == SERVER_CLIENT_STATE_DISCONNECTED) continue;
    clientCount++;
  } while(i < SERVER_MAX_CLIENTS);
  return clientCount;
}

void serverStop() {
  assertIsMainThread("Server stop must be on main thread");
  if(SERVER.state == SERVER_STATE_STOPPED) return;

  consolePrint("Stopping server...");
  switch(SERVER.type) {
    case SERVER_TYPE_NETWORKED:
      networkedServerStop(&SERVER);
      break;

    default:
      assertUnreachable("Invalid server type");
  }
  consolePrint("Server stopped.");
}

void serverDispose() {
  serverStop();
}